/*
 *  bt2000e_cs: BlueZ HCI driver for pcmcia bluetooth cards which
 *  use Oxford Semi's OXCF950 UART/CF controller and Ericsson ROK 
 *  bluetooth module, such as Ambicom/Pretec BT2000E.
 *
 *  Copyright (C) 2002-2004  Albert Rybalkin <albertr@iral.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation;
 *
 *  Software distributed under the License is distributed on an "AS
 *  IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 *  implied. See the License for the specific language governing
 *  rights and limitations under the License.
 *
 * Portions based on the original code of btuart_cs.c by Marcel Holtmann,
 * Copyright (C) 2001-2002  Marcel Holtmann <marcel@holtmann.org>. 
 *
 * Portions based on the original code of pcmcia-cs by David A. Hinds,
 * Copyright (C) 1999 David A. Hinds.  All Rights Reserved.
 *
 */

/* 
   Principles of operations and notes on locking:
   =============================================
   
   To take advantage of automatic FIFO level triggers, we have both TX & RX 
   interrupts enabled and served by our hardware interrupt handler. RX is
   relatively quick and simple - it just reads available data up from 
   RX FIFO, no waiting or inter-process synchronization is involved. The
   only locking required is a general hardware lock (to make sure that
   hardware is not gone while we are using it), so simple read lock is
   enuff. Since RX is completely interrupt-driven and because of technical
   requirement for clearing RX interrupt (we need to perform an RX FIFO read
   to clear interrupt), it was decided to run RX completely in hardware
   interrupt context.
   
   On the other side, TX is both interrupt and protocol stack -driven. Also,
   TX could take awhile and may require an exclusive locking of hardware (in
   case of changing uart clock frequency divisor, for instance). To take
   advantage of in-kernel synchronization and scheduling, TX was implemented
   as a tasklet. It also simplifies locking, since we can now rely on 
   in-kernel serialization of tasklet execution. In theory, that also should
   allow us to process both TX and RX requests mostly simultaneously on SMP
   systems. But since tasklet runs in software interrupt context, it can't
   just sleep while waiting for Ericsson ROK module uart rate change. So a 
   timer was utilized to complete this wait and to optionally change OXCF950 
   uart baud rate, that timer requires an exclusive lock. 
   
   Thus, the only lock we use is rwlock for direct hardware access. Any non-
   destructive operation needs to get a read lock, but destructive ones 
   (stop/start hardware, change clock divisor, etc) require an exclusive 
   write lock and disabling interrupts on local CPU to avoid deadlocks.
*/

#include <linux/config.h>

#ifdef CONFIG_MODVERSIONS
#ifndef MODVERSIONS
#define MODVERSIONS
#endif
#include <linux/modversions.h>
#endif

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/types.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/errno.h>
#include <linux/ptrace.h>
#include <linux/ioport.h>
#include <linux/spinlock.h>
#include <linux/delay.h>

#include <linux/skbuff.h>
#include <linux/string.h>
#include <linux/serial.h>
#include <linux/serial_reg.h>
#include <asm/system.h>
#include <asm/bitops.h>
#include <asm/io.h>

#include <pcmcia/version.h>
#include <pcmcia/cs_types.h>
#include <pcmcia/cs.h>
#include <pcmcia/cistpl.h>
#include <pcmcia/ciscode.h>
#include <pcmcia/ds.h>
#include <pcmcia/cisreg.h>

#include <net/bluetooth/bluetooth.h>
#include <net/bluetooth/hci_core.h>

/* ======================== Module parameters ======================== */

#define DRIVER_VERSION 	"0.28"
#define DRIVER_NAME	"bt2000e_cs: BT2000E BlueZ HCI pcmcia driver"
#define OX950_IRQ
#define OX950_HCIRST
//#define OX950_CORRST
//#define OX950_DEBUG

static int baud_rate = 460800;	/* default baud rate to switch to */
static int divisor   = 1;	/* uart clock divisor for this rate */ 	

/* Bit map of interrupts to choose from */
static u_int irq_mask = 0x86bc;
static int irq_list[4] = { -1 };

MODULE_PARM(irq_mask, "i");
MODULE_PARM(irq_list, "1-4i");
MODULE_PARM(baud_rate, "i");
MODULE_PARM_DESC(baud_rate, "Use the specific baud rate from the following list: 460800,230400,153600,115200,76800,57600,38400,28800,19200,9600");
MODULE_AUTHOR("Albert Rybalkin <albertr@iral.com>");
MODULE_DESCRIPTION(DRIVER_NAME DRIVER_VERSION);
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif

/* ======================== Local structures ======================== */

typedef struct bt2000e_info_t {

    dev_link_t link;
    dev_node_t node;

    struct hci_dev hdev;

    rwlock_t lock;			/* serializing access to hardware */

    struct tasklet_struct tx_task;	/* TX worker */
    struct timer_list rst_timer;	/* change baud rate wait event */
    struct sk_buff_head txq;

    unsigned long tx_state;

    unsigned int rx_count;
    unsigned int rx_state;
    
    struct sk_buff *rx_skb;

} bt2000e_info_t;

/* Here's the list of baud rates supported by ericsson ROK module
   and corresponding OXCF950's clock divisors. 
   
   Note: not all combinations were tested, some may not work. 
   
   Our default rate is 460800 bps. To change it, use baud_rate parameter.

   According to Ericsson, the ROK module can take up to 15ms to switch
   the rate and if it's already running slower than 9600 bps, then it's
   possible that it will change the rate just before all bytes of command
   complete event have been sent out. Because of that, we don't want to go
   below 9600 bps.
*/
#if defined(OX950_HCIRST)
static const struct {
        int baud_rate;
        int divisor;
        unsigned char cmd;
} baud_map[] = {
/*
baud_rate	divisor	command
=========	======= ======= */
{460800,	1,	0x00}, /* our default setting */
{230400,	2,	0x01},
{153600,	3,	0x10},
{115200,	4,	0x02},
{76800,		6,	0x11},
{57600,		8,	0x03}, /* ericsson default setting */
{38400,		12,	0x12},
{28800,		16,	0x04},
{19200,		24,	0x13},
{9600,		48,	0x14},
};

#define BAUD_MAP_SIZE ARRAY_SIZE(baud_map)

/* HCI reset & ericsson set uart baud rate commands */
static unsigned char hci_cmd[] = { 
			HCI_COMMAND_PKT, 0x03, 0x0C, 0x00, /* reset */
			HCI_COMMAND_PKT, 0x09, 0xFC, 0x01, 0x00 /* set ericsson uart baud rate */
};
#endif

static dev_info_t dev_info = "bt2000e_cs";
static dev_link_t *dev_list = NULL;

static void bt2000e_config(dev_link_t *link);
static void bt2000e_release(u_long arg);
static int bt2000e_event(event_t event, int priority, event_callback_args_t *args);
static dev_link_t *bt2000e_attach(void);
static void bt2000e_detach(dev_link_t *);


/* Receiver states */
#define RECV_WAIT_PACKET_TYPE   0
#define RECV_WAIT_EVENT_HEADER  1
#define RECV_WAIT_ACL_HEADER    2
#define RECV_WAIT_SCO_HEADER    3
#define RECV_WAIT_DATA          4

/* Transmitter states */
#define XMIT_FRAG	0 /* set if we are in the middle of skb transmittion */
#define XMIT_UART_DIV	1 /* set if we need to change UART divisor */
	
/* 950-specific stuff */
#define MAX_WAIT	0xFFFF
#define FIFO_SIZE	128

#define TR_TX_INT 	0x10	/* TTL: TX interrupt trigger level (0-127) */
#define TR_RX_INT 	0x40	/* RTL: RX interrupt trigger level (1-127) */
#define TR_CTL_LO 	0x08	/* FCL: auto flow control LOWER trigger level (0-127) */
#define TR_CTL_HI 	0x60	/* FCH: auto flow control HIGH trigger level (1-127) */

/* 950-specific registers and values we use. It should eventually go to
 *
 *  include/linux/serial_reg.h
 *
 */
#define UART_IER_CTS	0x80 /* enable CTS interrupt */
#define UART_IER_RTS	0x40 /* enable RTS interrupt */
#define UART_IER_SLP	0x10 /* enable sleep mode */
#define UART_LCR_650	0xBF /* enable 650-compatible registers access */
#define UART_LSR_DE	0x80 /* data error */
#define UART_IIR_RXTOUT 0x0C /* RX timeout */
#define UART_IIR_CTSRTS 0x20 /* CTS or RTS change interrupt */
#define UART_IIR_RTS	0x40
#define UART_IIR_CTS	0x80
#define UART_IIR_MASK	0x3E /* interrupt mask */
#define UART_SRT	0x0D /* soft reset register */

/* ======================== Interrupt handling ======================== */

/* a timer to be schedultd by rx tasklet to complete wait for
   ericsson ROK module's baud rate change. Note: rx tasklet is
   disabled duting the waiting, interrupts are enabled. 
   Requires an exclusive lock while changing uart clock divisor */
#if defined(OX950_HCIRST)
static void bt2000e_rst_wakeup(unsigned long arg) { 

    bt2000e_info_t *info = (bt2000e_info_t*) arg;

    if (!info) {
	printk(KERN_ERR "bt2000e_cs: unknown device\n");
	return;
    }

    if (test_bit(XMIT_UART_DIV, &(info->tx_state))) {
    
	unsigned long flags;
	unsigned int iobase;
	unsigned char lcr;
	
	/* we need an exclusive lock */
	write_lock_irqsave(&(info->lock), flags);

	if (!(info->link.state & DEV_CONFIG)) {
	    printk(KERN_ERR "bt2000e_cs: hardware is gone?\n");
	    write_unlock_irqrestore(&(info->lock), flags);
	    return;
	}

	iobase = info->link.io.BasePort1;
	
	lcr = inb(iobase + UART_LCR) & 0x3F; /* mask out UART_LCR_DLAB and UART_LCR_SBC */
    
	/* enable divisor latch access */
	outb(lcr | UART_LCR_DLAB, iobase + UART_LCR);

	/* setup divisor latch */
	outb(divisor & 0x00FF, iobase + UART_DLL); /* divisor latch LOW byte */
	outb((divisor & 0xFF00) >> 8, iobase + UART_DLM); /* divisor latch HIGH byte */

	/* disable divisor latch access */
	outb(lcr, iobase + UART_LCR);

#if defined(OX950_DEBUG)
	printk(KERN_DEBUG "bt2000e_cs: new uart clock divisor = %d\n", divisor);
#endif
	/* no need to change it again */
	clear_bit(XMIT_UART_DIV, &(info->tx_state));

	write_unlock_irqrestore(&(info->lock), flags);
    
    }
    
    tasklet_enable(&(info->tx_task));

}
#endif

/* a tasklet to be scheduled by both hardware interrupt
   handler and protocol stack. It could also re-schedule itself. Requires
   a read lock to make sure hardware is sane */
static void bt2000e_tx(unsigned long arg) {

    bt2000e_info_t *info = (bt2000e_info_t*) arg;
    unsigned int iobase;
    register int i;
    struct sk_buff *skb = NULL;
    int size, space_avail, space_left;

    if (!info) {
	printk(KERN_ERR "bt2000e_cs: unknown device\n");
	return;
    }

    read_lock(&(info->lock));

    if (!(info->link.state & DEV_CONFIG)) {
	printk(KERN_ERR "bt2000e_cs: hardware is gone?\n");
	read_unlock(&(info->lock));
	return;
    }

    iobase = info->link.io.BasePort1;
	
    if (skb_queue_empty(&(info->txq)) || !(skb = skb_dequeue(&(info->txq)))) {
#if defined(OX950_DEBUG)
	printk(KERN_DEBUG "bt2000e_cs: empty TX queue\n");
#endif
	clear_bit(XMIT_FRAG, &(info->tx_state));
	read_unlock(&(info->lock));
	return;
    }

    /* activate DTR and RTS */
    outb(UART_MCR_DTR | UART_MCR_RTS | UART_MCR_OUT2, iobase + UART_MCR);

    /* wait for CTS */
    for (i = MAX_WAIT; i; i--)
	if (inb(iobase + UART_MSR) & UART_MSR_CTS)
	    break;

    if (!i) {
	printk(KERN_ERR "bt2000e_cs: timeout waiting for CTS\n");
	/* drop skb on the floor */
	dev_kfree_skb_any(skb);
	clear_bit(XMIT_FRAG, &(info->tx_state));	    
	read_unlock(&(info->lock));
	return;
    }
    
    space_avail = FIFO_SIZE; /* when scheduled by the interrupt handler, 
			    we might actually have less than that available,
			    but we're gonna wait for it to drain, anyway */

    /* TX FIFO should be empty */
    for (i = MAX_WAIT; i; i--)
        if (inb(iobase + UART_LSR) & UART_LSR_THRE)
	    break;

    if (!i) {
        printk(KERN_ERR "bt2000e_cs: timeout waiting for empty TX FIFO\n");
        /* drop skb on the floor */
        dev_kfree_skb_any(skb);
        clear_bit(XMIT_FRAG, &(info->tx_state));
        read_unlock(&(info->lock));
        return;
    }
        
start_over:

    size = skb->len;

    /* Check for HCI_Reset command. On reset, ericsson ROK module
       defaults to 57600 baud, we need to switch it back to 
       our selected baud rate. We only check for HCI_Reset command 
       at the begining position of skb.
    */
#if defined(OX950_HCIRST)
    if (!test_bit(XMIT_FRAG, &(info->tx_state)) &&
	(size >= 4) &&
	(skb->data[0] == 0x01) && 
	(skb->data[1] == 0x03) && 
	(skb->data[2] == 0x0C) && 
	(skb->data[3] == 0x00)) {
	
	    if (sizeof(hci_cmd) > space_avail) {
		/* don't have enuff free space in TX FIFO left, bail out */
		skb_queue_head(&(info->txq), skb);
		clear_bit(XMIT_FRAG, &(info->tx_state));	    
		tasklet_schedule(&(info->tx_task));
		read_unlock(&(info->lock));
		return;
	    }
	    
	    /* fill up TX FIFO */
	    for (i = 0; i < sizeof(hci_cmd); i++)
		outb(hci_cmd[i], iobase + UART_TX);
	    
	    space_avail -= sizeof(hci_cmd);
	    info->hdev.stat.byte_tx += sizeof(hci_cmd);
	    
	    size -= 4;

	    if (size) {
		/* we have more data in the skb */
		skb_pull(skb, 4);
		skb_queue_head(&(info->txq), skb);
		/* we don't expect to find two or more HCI_Reset 
		commands in one skb */
		set_bit(XMIT_FRAG, &(info->tx_state));
		tasklet_schedule(&(info->tx_task));
	    } else {
		dev_kfree_skb_any(skb);
		skb = NULL;
		clear_bit(XMIT_FRAG, &(info->tx_state));	    
	    }
	    
	    printk(KERN_NOTICE "bt2000e_cs: hci_reset\n");

	    /* According to Ericsson, the ROK module could take up to 15ms
	    to change its uart rate *AFTER* sending out command complete 
	    event. Possible protocol issue: what protocol stack will be 
	    doing with command complete events for commands it didn't send? */

	    tasklet_disable(&(info->tx_task));

	    init_timer(&(info->rst_timer));
	    info->rst_timer.function = bt2000e_rst_wakeup;
	    info->rst_timer.data = (unsigned long) info;
	    info->rst_timer.expires = jiffies + HZ/50; /* ~20ms */
	    add_timer(&(info->rst_timer));

	    read_unlock(&(info->lock));
	    return;
    }
#endif

    if (size > space_avail) {
	
	space_left = 0;
	
	/* fill up TX FIFO */
	for (i = 0; i < space_avail; i++)
	    outb(skb->data[i], iobase + UART_TX);

	/* put unsent data back to the queue */	    
	skb_pull(skb, space_avail);
	skb_queue_head(&(info->txq), skb);
	info->hdev.stat.byte_tx += space_avail;
	set_bit(XMIT_FRAG, &(info->tx_state));
	tasklet_schedule(&(info->tx_task));

    } else {

	space_left = space_avail - size;
	    
	/* fill up TX FIFO */
	for (i = 0; i < size; i++)
	    outb(skb->data[i], iobase + UART_TX);
	    
	dev_kfree_skb_any(skb);
	skb = NULL;
	info->hdev.stat.byte_tx += size;
	clear_bit(XMIT_FRAG, &(info->tx_state));
		
    }

#if defined(OX950_DEBUG)
	printk(KERN_DEBUG "bt2000e_cs: wrote=<%d out of %d>\n", (size > space_avail) ? space_avail : size, size);
#endif

#if defined(OX950_IRQ)
    if (space_left) {
	/* we still have some extra space in TX FIFO, try to pull out 
	    another skb from the TX queue, if any available */
	if (!skb_queue_empty(&(info->txq)) && (skb = skb_dequeue(&(info->txq)))) {
	    space_avail = space_left;
	    goto start_over;
	}
    }
#endif

    read_unlock(&(info->lock));

}

/* here's an ugly trick we use to ping Ericsson ROK module
   to bring it back from coma after power up. The caller
   should already hold a lock */
static inline void bt2000e_tx_cts(bt2000e_info_t *info) {

    unsigned int iobase;
    register int i;
    unsigned long flags;
            
    if (!info) {
	printk(KERN_ERR "bt2000e_cs: unknown device\n");
	return;
    }

    write_lock_irqsave(&(info->lock), flags);

    if (!(info->link.state & DEV_CONFIG)) {
	printk(KERN_ERR "bt2000e_cs: hardware is gone?\n");
	write_unlock_irqrestore(&(info->lock), flags);
	return;
    }
    
    iobase = info->link.io.BasePort1;

    /* activate DTR and RTS */
    outb(UART_MCR_DTR | UART_MCR_RTS | UART_MCR_OUT2, iobase + UART_MCR);

    /* wait for CTS */
    for (i = MAX_WAIT; i; i--)
	if (inb(iobase + UART_MSR) & UART_MSR_CTS)
	    break;

    write_unlock_irqrestore(&(info->lock), flags);

}

/* bt2000e_rx_wakeup is only called by hardware interrupt handler
   that already holds a read lock. Since it's executed in hardware
   interrupt context, it needs to return as soon as possible */ 
static inline void bt2000e_rx_wakeup(bt2000e_info_t *info) {

    unsigned int iobase = info->link.io.BasePort1;
    hci_event_hdr *eh;
    hci_acl_hdr   *ah;
    hci_sco_hdr   *sh;
    register unsigned int size = 0;

start_over:

    /* Allocate packet if needed */
    if (!(info->rx_skb) && !(info->rx_skb = bluez_skb_alloc(HCI_MAX_FRAME_SIZE, GFP_ATOMIC))) {
        printk(KERN_ERR "bt2000e_cs: can't allocate mem for new packet\n");
	info->rx_skb = NULL;
	/* clear interrupt */
	inb(iobase + UART_RX);
	return;
    }

    if (info->rx_state == RECV_WAIT_PACKET_TYPE) {

	inb(iobase + UART_MCR);

	/* get first byte */
	info->rx_skb->pkt_type = inb(iobase + UART_RX);

	info->rx_skb->dev = (void *)&(info->hdev);
	info->hdev.stat.byte_rx++;

	switch (info->rx_skb->pkt_type) {

    	    case HCI_EVENT_PKT:
    		info->rx_state = RECV_WAIT_EVENT_HEADER;
    		info->rx_count = HCI_EVENT_HDR_SIZE;
    		break;

	    case HCI_ACLDATA_PKT:
    		info->rx_state = RECV_WAIT_ACL_HEADER;
    		info->rx_count = HCI_ACL_HDR_SIZE;
    		break;

	    case HCI_SCODATA_PKT:
    		info->rx_state = RECV_WAIT_SCO_HEADER;
    		info->rx_count = HCI_SCO_HDR_SIZE;
    		break;

	    default:
    		/* bogus packet, read the rest and discard it */
		size = 1;
    		info->hdev.stat.err_rx++;

		while (inb(iobase + UART_LSR) & UART_LSR_DR) {
		    inb(iobase + UART_RX);
		    size++;
		}
		
    		printk(KERN_ERR "bt2000e_cs: received bogus packet: type=0x%02X size=%d\n", info->rx_skb->pkt_type, size);
		dev_kfree_skb_any(info->rx_skb);
		info->rx_skb = NULL;
    		return;
    		break;
	}
    
    }

get_more:

    if (info->rx_count > skb_tailroom(info->rx_skb)) {
	printk(KERN_ERR "bt2000e_cs: received packet is too large\n");
    	info->hdev.stat.err_rx++;
	dev_kfree_skb_any(info->rx_skb);
	info->rx_skb = NULL;
	/* clear interrupt */
	inb(iobase + UART_RX);
	return;
    }

    if (!(info->rx_count)) {
	printk(KERN_ERR "bt2000e_cs: received packet is zero size\n");
    	info->hdev.stat.err_rx++;
	dev_kfree_skb_any(info->rx_skb);
	info->rx_skb = NULL;
	/* clear interrupt */
	inb(iobase + UART_RX);
	return;
    }

    size = info->rx_count;

    while ((info->rx_count) && (inb(iobase + UART_LSR) & UART_LSR_DR)) {

	/* get the next byte */
	*skb_put(info->rx_skb, 1) = inb(iobase + UART_RX);
    	info->hdev.stat.byte_rx++;
    	info->rx_count--;
    }

#if defined(OX950_DEBUG)
	printk(KERN_DEBUG "bt2000e_cs: received type=0x%X state=%d size=<%d out of %d>\n", info->rx_skb->pkt_type, info->rx_state, size - info->rx_count, size);
#endif

    if (!(info->rx_count)) {
        switch (info->rx_state) {

	    case RECV_WAIT_EVENT_HEADER:
	        eh = (hci_event_hdr *)(info->rx_skb->data);
		info->rx_state = RECV_WAIT_DATA;
		info->rx_count = eh->plen;
		goto get_more;
		break;

	    case RECV_WAIT_ACL_HEADER:
		ah = (hci_acl_hdr *)(info->rx_skb->data);
		info->rx_state = RECV_WAIT_DATA;
		info->rx_count = __le16_to_cpu(ah->dlen);
		goto get_more;
		break;

    	    case RECV_WAIT_SCO_HEADER:
		sh = (hci_sco_hdr *)(info->rx_skb->data);
		info->rx_state = RECV_WAIT_DATA;
		info->rx_count = sh->dlen;
		goto get_more;
		break;

    	    case RECV_WAIT_DATA:
		hci_recv_frame(info->rx_skb);
		info->rx_state = RECV_WAIT_PACKET_TYPE;
		info->rx_skb = NULL;
#if defined(OX950_IRQ)
		if (inb(iobase + UART_LSR) & UART_LSR_DR)
		    /* we can save an interrupt here */
		    goto start_over;
#endif
		break;
	}
    }

    return;
}


/* hardware interrupt handler. Needs a read lock to make sure that hardware
   is sane. */
static void bt2000e_interrupt(int irq, void *dev_inst, struct pt_regs *regs) {

    bt2000e_info_t *info = dev_inst;
    unsigned int iobase;
    unsigned char lsr, msr;
    register unsigned char iir;

    if (!info) {
	printk(KERN_ERR "bt2000e_cs: unknown device for irq %d\n", irq);
	return;
    }

    /* kernel guarantees that this handler is never re-entered,
    can use a read lock which does not block other readers... */
    read_lock(&(info->lock));

    if (!(info->link.state & DEV_CONFIG)) {
	printk(KERN_ERR "bt2000e_cs: hardware is gone?\n");
	read_unlock(&(info->lock));
	return;
    }
    
    iobase = info->link.io.BasePort1;

    iir = inb(iobase + UART_IIR);

    while (!(iir & UART_IIR_NO_INT)) {

	switch (iir & UART_IIR_MASK) {

            case UART_IIR_RDI:
		/* RX interrupt */
#if defined(OX950_DEBUG)
	        printk(KERN_DEBUG "bt2000e_cs: RX interrupt\n");
#endif
		bt2000e_rx_wakeup(info);
		break;

	    case UART_IIR_RXTOUT:
		/* RX timeout interrupt */
#if defined(OX950_DEBUG)
		printk(KERN_DEBUG "bt2000e_cs: RX timeout interrupt\n");
#endif
		bt2000e_rx_wakeup(info);
		break;
	
    	    case UART_IIR_THRI:
		/* TX interrupt */
#if defined(OX950_DEBUG)
		printk(KERN_DEBUG "bt2000e_cs: TX interrupt\n");
#endif
		/* writing to the card could be time
		consuming, schedule tasklet to complete the bottom half */
		tasklet_schedule(&(info->tx_task));
		break;

    	    case UART_IIR_RLSI:
		/* clear interrupt */
		lsr = inb(iobase + UART_LSR); 
		printk(KERN_ERR "bt2000e_cs: unhandled RLSI interrupt, LSR=0x%X\n", lsr);
		/* FIXME: we need to process error... */
		break;

	    case UART_IIR_MSI:
		/* clear interrupt */
		msr = inb(iobase + UART_MSR);
		printk(KERN_ERR "bt2000e_cs: unhandled MSI interrupt, MSR=0x%X\n", msr);
		/* FIXME: we need to process error... */
		break;

	    default:
		/* unhandled interrupt */
		printk(KERN_ERR "bt2000e_cs: unhandled interrupt IIR=0x%X\n", iir);
		/* FIXME: it should never happen, WTF???? */
		break;
	}
	
	iir = inb(iobase + UART_IIR);

    }
    
    read_unlock(&(info->lock));
    return;
}


/* ======================== HCI interface ======================== */


static int bt2000e_hci_send_frame(struct sk_buff *skb) {

    struct hci_dev* hdev = (struct hci_dev *)(skb->dev);
    bt2000e_info_t *info;

    if (!skb) {
	printk(KERN_ERR "bt2000e_cs: invalid skb\n");
	return -ENODEV;
    }

    if (!hdev) {
	printk(KERN_ERR "bt2000e_cs: unknown HCI device\n");
	dev_kfree_skb_any(skb);
	return -ENODEV;
    }

    info = (bt2000e_info_t *)(hdev->driver_data);

    if (!(info->link.state & DEV_CONFIG)) {
	printk(KERN_ERR "bt2000e_cs: hardware is gone?\n");
	return -ENODEV;
    }

    switch (skb->pkt_type) {

	case HCI_COMMAND_PKT:

	    hdev->stat.cmd_tx++;
	    break;

	case HCI_ACLDATA_PKT:
	    
	    hdev->stat.acl_tx++;
	    break;

	case HCI_SCODATA_PKT:
    
	    hdev->stat.sco_tx++;
	    break;
    };

    if (!skb_headroom(skb)) {
	printk(KERN_ERR "bt2000e_cs: no room to extend skb\n");
	dev_kfree_skb_any(skb);
	return -ENODEV;
    }
    
    /* prepend skb with a frame type */
    memcpy(skb_push(skb, 1), &(skb->pkt_type), 1);
    skb_queue_tail(&(info->txq), skb);

    tasklet_schedule(&(info->tx_task));
    
    return 0;

}


static void bt2000e_hci_reset(bt2000e_info_t *info) {
    
    struct hci_dev *hdev = &(info->hdev);
    struct sk_buff *skb;
    
    unsigned char cmd[] = {HCI_COMMAND_PKT, 0x03, 0x0c, 0x00}; /* HCI_Reset */

    if (!hdev) {
	printk(KERN_ERR "bt2000e_cs: unknown HCI device\n");
	return;
    }

    if (!(info->link.state & DEV_CONFIG)) {
	printk(KERN_ERR "bt2000e_cs: hardware is gone?\n");
	return;
    }

    if (!(skb = bluez_skb_alloc(HCI_MAX_FRAME_SIZE, GFP_ATOMIC))) {
	printk(KERN_ERR "bt2000e_cs: can't allocate mem for new packet\n");
	return;
    }
			    
    memcpy(skb_put(skb, sizeof(cmd)), cmd, sizeof(cmd));

    hdev->stat.cmd_tx++;

    skb_queue_tail(&(info->txq), skb);

//    tasklet_schedule(&(info->tx_task));

    return;
}


static int bt2000e_hci_flush(struct hci_dev *hdev) {

    bt2000e_info_t *info = (bt2000e_info_t *)(hdev->driver_data);

    /* Drop TX queue & RX skb */
    skb_queue_purge(&(info->txq));

    if (info->rx_skb) {
	dev_kfree_skb_any(info->rx_skb);
	info->rx_skb = NULL;
    }

    return 0;

}


static int bt2000e_hci_open(struct hci_dev *hdev) {

    bt2000e_info_t *info = (bt2000e_info_t *)(hdev->driver_data);

    /* reset */
    bt2000e_hci_reset(info);

    tasklet_schedule(&(info->tx_task));

    set_bit(HCI_RUNNING, &(hdev->flags));

    return 0;

}


static int bt2000e_hci_close(struct hci_dev *hdev) {

    if (!test_and_clear_bit(HCI_RUNNING, &(hdev->flags)))
	return 0;

    bt2000e_hci_flush(hdev);

    return 0;

}


static void bt2000e_hci_destruct(struct hci_dev *hdev) {

    /* nothing to do */
  
}


static int bt2000e_hci_ioctl(struct hci_dev *hdev, unsigned int cmd, unsigned long arg) {

    /* we don't provide any ioctl's */
    return -ENOIOCTLCMD;

}


/* ======================== Card services HCI interaction ======================== */

#if defined(OX950_CORRST)
static void pcmcia_cor_reset(bt2000e_info_t *info) {

    conf_reg_t reg;
    int err;

    if (!info) {
	printk(KERN_ERR "bt2000e_cs: unknown device\n");
	return;
    }

    reg.Function = 0;
    reg.Action = CS_READ;
    reg.Offset = CISREG_COR;
    reg.Value = 0;
    
    err = pcmcia_access_configuration_register(info->link.handle, &reg);
    
    if (err != CS_SUCCESS) {
        printk(KERN_DEBUG "bt2000e_cs: COR reset failed (%d)\n", err);
        return;
    }

#if defined(OX950_DEBUG)
    printk(KERN_DEBUG "bt2000e_cs: original COR %02x\n", reg.Value);
#endif

    reg.Action = CS_WRITE;
    reg.Value |= COR_SOFT_RESET;
    
    err = pcmcia_access_configuration_register(info->link.handle, &reg);
    
    if (err != CS_SUCCESS) {
        printk(KERN_DEBUG "bt2000e_cs: COR reset failed (%d)\n", err);
        return;
    }

    mdelay(10);

    reg.Value &= ~COR_SOFT_RESET;
    
    err = pcmcia_access_configuration_register(info->link.handle, &reg);
        
    if (err != CS_SUCCESS) {
        printk(KERN_DEBUG "bt2000e_cs: COR reset failed (%d)\n", err);
        return;
    }

    mdelay(10);
}
#endif

static int bt2000e_start_uart(bt2000e_info_t *info) {

    unsigned int iobase = info->link.io.BasePort1;
    unsigned char lcr, ier = UART_IER_THRI | UART_IER_RDI | UART_IER_RLSI | UART_IER_SLP;
    unsigned int defdivisor = 8; /* 57600 baud is default rate on ericsson module */
    unsigned char id1, id2, id3, rev;
    register int i;
    
    /* disable interrupts */
    outb(0, iobase + UART_IER); 

    outb(UART_MCR_RTS | UART_MCR_OUT2, iobase + UART_MCR);

    /* setup FIFOs */
    outb(0, iobase + UART_FCR);
    inb(iobase + UART_RX);
    outb(UART_FCR_ENABLE_FIFO | UART_FCR_CLEAR_RCVR | UART_FCR_CLEAR_XMIT | UART_FCR_TRIGGER_14, iobase + UART_FCR);
    
    /* disable divisor latch access */
    lcr = inb(iobase + UART_LCR) & 0x3F; /* mask out UART_LCR_DLAB and UART_LCR_SBC */
    outb(lcr, iobase + UART_LCR);
    
    /* read upto 4 bytes from RX FIFO */
    for(i = 1; i < 5; i++) {
        inb(iobase + UART_RX);
        if (!(inb(iobase + UART_LSR) & UART_LSR_DR))
	    break;
    }

    /* wait if CTS/DSR/DCD changing */
    for (i = 1; i < 0x3E8; i++) {
        if (!(inb(iobase + UART_MSR) & UART_MSR_ANY_DELTA))
	    break;
    }
    
    /* enable divisor latch access */
    outb(lcr | UART_LCR_DLAB, iobase + UART_LCR);

#if defined(OX950_DEBUG)
	printk(KERN_DEBUG "bt2000e_cs: default uart clock divisor = %d\n", defdivisor);
#endif

    /* setup divisor latch */
    outb(defdivisor & 0x00FF, iobase + UART_DLL); /* divisor latch LOW byte */
    outb((defdivisor & 0xFF00) >> 8, iobase + UART_DLM); /* divisor latch HIGH byte */

    /* disable divisor latch access */
    outb(lcr, iobase + UART_LCR);

    /* setup interrupts, enable sleep mode */
    outb(ier, iobase + UART_IER);

    /* skip any pending interrupts */
    for (i = 1; i < 5; i++) {
        if (inb(iobase + UART_IIR) & UART_IIR_NO_INT)
	    break;
    }

    /* 8N1 */
    lcr = UART_LCR_WLEN8;
    outb(lcr, iobase + UART_LCR); 

    /* setup CTS/RTS flow control and 950 enhanced mode */
    outb(UART_LCR_650, iobase + UART_LCR);
    outb(UART_EFR_CTS | UART_EFR_RTS | UART_EFR_ECB, iobase + UART_EFR);
    outb(lcr, iobase + UART_LCR);

    /* read core id and revision */
    outb(UART_ACR, iobase + UART_EMSR);
    outb(UART_ACR_ICRRD, iobase + UART_LSR); /* enable ICR read access, we don't need to save the old value of ACR */

    outb(UART_ID1, iobase + UART_EMSR);
    id1 = inb(iobase + UART_LSR);

    outb(UART_ID2, iobase + UART_EMSR);
    id2 = inb(iobase + UART_LSR);

    outb(UART_ID3, iobase + UART_EMSR);
    id3 = inb(iobase + UART_LSR);

    outb(UART_REV, iobase + UART_EMSR);
    rev = inb(iobase + UART_LSR);

    if (id1 != 0x16 || id2 != 0xC9 || id3 != 0x50) {
        printk(KERN_ERR "bt2000e_cs: unknown UART core: 0x%02x%02x%02x\n", id1, id2, id3);
        return -ENODEV;
    }

#if defined(OX950_DEBUG)
    switch (rev) {

        case 3:
	    printk(KERN_DEBUG "bt2000e_cs: core is OX16C950B\n");
    	    break;

	case 6:
    	    printk(KERN_DEBUG "bt2000e_cs: core is OXCF950\n");
    	    break;

	case 8:
    	    printk(KERN_DEBUG "bt2000e_cs: core is OXCF950B\n");
    	    break;

	default:
    	    printk(KERN_DEBUG "bt2000e_cs: core revision: 0x%X\n", rev);
    	    break;
    }
#endif

    /* init ICR registers */
    outb(UART_TTL, iobase + UART_EMSR);
    outb(TR_TX_INT, iobase + UART_LSR); /* TX interrupt trigger level (0-127) */

    outb(UART_RTL, iobase + UART_EMSR);
    outb(TR_RX_INT, iobase + UART_LSR); /* RX interrupt trigger level (1-127) */

    outb(UART_FCL, iobase + UART_EMSR);
    outb(TR_CTL_LO, iobase + UART_LSR); /* auto flow control LOWER trigger level (0-127) */

    outb(UART_FCH, iobase + UART_EMSR);
    outb(TR_CTL_HI, iobase + UART_LSR); /* auto flow control HIGH trigger level (1-127) */

    outb(UART_ACR, iobase + UART_EMSR);
    outb(UART_ACR_TLENB, iobase + UART_LSR); /* disable ICR read access, enable trigger levels */ 

#if defined(OX950_HCIRST)
    for (i = 0; i < BAUD_MAP_SIZE; i++) {
	if (baud_rate == baud_map[i].baud_rate) {
	    divisor = baud_map[i].divisor;
	    hci_cmd[8] = baud_map[i].cmd;
	    if (baud_rate != 57600) /* ericsson default */
		set_bit(XMIT_UART_DIV, &(info->tx_state));
	    printk(KERN_NOTICE "bt2000e_cs: requested baud rate = %d bps\n", baud_rate);
	    break;
	}
    }

    if (i == BAUD_MAP_SIZE ) {
	printk(KERN_NOTICE "bt2000e_cs: requested invalid baud rate %d bps\n", baud_rate);
	baud_rate = baud_map[0].baud_rate; /* our default */
	divisor = baud_map[0].divisor;
	hci_cmd[8] = baud_map[0].cmd;
	set_bit(XMIT_UART_DIV, &(info->tx_state));
	printk(KERN_NOTICE "bt2000e_cs: set default baud rate %d bps\n", baud_rate);
    }
#endif
    
    return 0;

}


static void bt2000e_stop_uart(bt2000e_info_t *info) {

    unsigned int iobase = info->link.io.BasePort1;

    /* disable interrupts */
    outb(0, iobase + UART_IER);
    outb(0, iobase + UART_MCR);

    return;
}


static int bt2000e_open(bt2000e_info_t *info) {

    struct hci_dev *hdev;
    int ret;
    unsigned long flags;

    rwlock_init(&(info->lock));

    skb_queue_head_init(&(info->txq));

    info->rx_state = RECV_WAIT_PACKET_TYPE;
    info->rx_count = 0;
    info->rx_skb = NULL;
    clear_bit(XMIT_FRAG, &(info->tx_state));
    clear_bit(XMIT_UART_DIV, &(info->tx_state));

    write_lock_irqsave(&(info->lock), flags);

    /* start hardware */
    if ((ret = bt2000e_start_uart(info)) != 0) {
	write_unlock_irqrestore(&(info->lock), flags);
	return ret;
    }
    
    /* Initialize and register HCI device */
    hdev = &(info->hdev);

    hdev->type = HCI_PCCARD;
    hdev->driver_data = info;

    hdev->open     = bt2000e_hci_open;
    hdev->close    = bt2000e_hci_close;
    hdev->flush    = bt2000e_hci_flush;
    hdev->send     = bt2000e_hci_send_frame;
    hdev->destruct = bt2000e_hci_destruct;
    hdev->ioctl    = bt2000e_hci_ioctl;

    tasklet_init(&(info->tx_task), (void*)bt2000e_tx, (unsigned long) info);
    
    write_unlock_irqrestore(&(info->lock), flags);

    /* wait before it's safe to access hardware */
    set_current_state(TASK_INTERRUPTIBLE);
    schedule_timeout(HZ);

    if (hci_register_dev(hdev) < 0) {
	printk(KERN_ERR "bt2000e_cs: can't register HCI device %s.\n", hdev->name);
	return -ENODEV;
    }

    return 0;

}


static int bt2000e_close(bt2000e_info_t *info) {

    struct hci_dev *hdev = &(info->hdev);
    unsigned long flags;

    bt2000e_hci_close(hdev);

    if (hci_unregister_dev(hdev) < 0)
	printk(KERN_ERR "bt2000e_cs: can't unregister HCI device %s.\n", hdev->name);

    write_lock_irqsave(&(info->lock), flags);

    /* stop hardware */
    bt2000e_stop_uart(info);

    tasklet_kill(&(info->tx_task));

    write_unlock_irqrestore(&(info->lock), flags);

    return 0;

}


/* ======================== Card services ======================== */


static void cs_error(client_handle_t handle, int func, int ret) {

    error_info_t err = { func, ret };

    CardServices(ReportError, handle, &err);

}


static dev_link_t *bt2000e_attach(void) {

    bt2000e_info_t *info;
    client_reg_t client_reg;
    dev_link_t *link;
    int i, ret;


    /* Create new info device */
    info = kmalloc(sizeof(*info), GFP_KERNEL);
    if (!info)
	return NULL;
    memset(info, 0, sizeof(*info));


    link = &info->link;
    link->priv = info;

    link->release.function = &bt2000e_release;
    link->release.data = (u_long)link;
    link->io.Attributes1 = IO_DATA_PATH_WIDTH_8;
    link->io.NumPorts1 = 8;
    link->irq.Attributes = IRQ_TYPE_EXCLUSIVE | IRQ_HANDLE_PRESENT;
    link->irq.IRQInfo1 = IRQ_INFO2_VALID | IRQ_LEVEL_ID;

    if (irq_list[0] == -1)
	link->irq.IRQInfo2 = irq_mask;
    else
	for (i = 0; i < 4; i++)
	    link->irq.IRQInfo2 |= 1 << irq_list[i];

    link->irq.Handler = bt2000e_interrupt;
    link->irq.Instance = info;

    link->conf.Attributes = CONF_ENABLE_IRQ;
    link->conf.IntType = INT_MEMORY_AND_IO;
    link->conf.Present = PRESENT_OPTION | PRESENT_STATUS | PRESENT_PIN_REPLACE | PRESENT_COPY;

    /* Register with Card Services */
    link->next = dev_list;
    dev_list = link;
    client_reg.dev_info = &dev_info;
    client_reg.Attributes = INFO_IO_CLIENT | INFO_CARD_SHARE;
    client_reg.EventMask =  CS_EVENT_CARD_INSERTION | CS_EVENT_CARD_REMOVAL |
                	    CS_EVENT_RESET_PHYSICAL | CS_EVENT_CARD_RESET |
            		    CS_EVENT_PM_SUSPEND | CS_EVENT_PM_RESUME;
    client_reg.event_handler = &bt2000e_event;
    client_reg.Version = 0x0210;
    client_reg.event_callback_args.client_data = link;

    ret = CardServices(RegisterClient, &link->handle, &client_reg);

    if (ret != CS_SUCCESS) {
	cs_error(link->handle, RegisterClient, ret);
	bt2000e_detach(link);
	return NULL;
    }

    return link;

}


static void bt2000e_detach(dev_link_t *link) {

    bt2000e_info_t *info = link->priv;
    dev_link_t **linkp;
    int ret;

    /* Locate device structure */
    for (linkp = &dev_list; *linkp; linkp = &(*linkp)->next)
	if (*linkp == link)
	    break;

    if (*linkp == NULL)
	return;

    del_timer(&link->release);

    if (link->state & DEV_CONFIG)
	bt2000e_release((u_long)link);

    if (link->handle) {

	ret = CardServices(DeregisterClient, link->handle);
	
	if (ret != CS_SUCCESS)
	    cs_error(link->handle, DeregisterClient, ret);
    }

    /* Unlink device structure, free bits */
    *linkp = link->next;

    kfree(info);

}


static int get_tuple(int fn, client_handle_t handle, tuple_t *tuple,
		     cisparse_t *parse) {

    int i;


    i = CardServices(fn, handle, tuple);

    if (i != CS_SUCCESS)
	return CS_NO_MORE_ITEMS;

    i = CardServices(GetTupleData, handle, tuple);

    if (i != CS_SUCCESS)
	return i;

    return CardServices(ParseTuple, handle, tuple, parse);

}


#define first_tuple(a, b, c) get_tuple(GetFirstTuple, a, b, c)
#define next_tuple(a, b, c) get_tuple(GetNextTuple, a, b, c)

static void bt2000e_config(dev_link_t *link) {

    static ioaddr_t base[4] = { 0x2f8, 0x3e8, 0x2e8, 0x0 };
    client_handle_t handle = link->handle;
    bt2000e_info_t *info = link->priv;
    tuple_t tuple;
    u_short buf[256];
    cisparse_t parse;
    cistpl_cftable_entry_t *cf = &parse.cftable_entry;
    config_info_t config;
    int i, j, try, last_ret, last_fn;


    tuple.TupleData = (cisdata_t *)buf;
    tuple.TupleOffset = 0;
    tuple.TupleDataMax = 255;
    tuple.Attributes = 0;

    /* Get configuration register information */
    tuple.DesiredTuple = CISTPL_CONFIG;
    last_ret = first_tuple(handle, &tuple, &parse);

    if (last_ret != CS_SUCCESS) {
	last_fn = ParseTuple;
	goto cs_failed;
    }
    
    link->conf.ConfigBase = parse.config.base;
    link->conf.Present = parse.config.rmask[0];

    /* Configure card */
    link->state |= DEV_CONFIG;
    i = CardServices(GetConfigurationInfo, handle, &config);
    link->conf.Vcc = config.Vcc;


    /* First pass: look for a config entry that looks normal. */
    tuple.TupleData = (cisdata_t *)buf;
    tuple.TupleOffset = 0; tuple.TupleDataMax = 255;
    tuple.Attributes = 0;
    tuple.DesiredTuple = CISTPL_CFTABLE_ENTRY;
    
    /* Two tries: without IO aliases, then with aliases */
    for (try = 0; try < 2; try++) {
	
	i = first_tuple(handle, &tuple, &parse);
    
	while (i != CS_NO_MORE_ITEMS) {

	    if (i != CS_SUCCESS)
		goto next_entry;

	    if (cf->vpp1.present & (1<<CISTPL_POWER_VNOM))
		link->conf.Vpp1 = link->conf.Vpp2 = cf->vpp1.param[CISTPL_POWER_VNOM]/10000;

	    if ((cf->io.nwin > 0) && (cf->io.win[0].len == 8) && (cf->io.win[0].base != 0)) {
		link->conf.ConfigIndex = cf->index;
		link->io.BasePort1 = cf->io.win[0].base;
		link->io.IOAddrLines = (try == 0) ? 16 : cf->io.flags & CISTPL_IO_LINES_MASK;
        
		i = CardServices(RequestIO, link->handle, &link->io);

		if (i == CS_SUCCESS)
		    goto found_port;
	    }
next_entry:

	    i = next_tuple(handle, &tuple, &parse);
	}
    }

    /* Second pass: try to find an entry that isn't picky about
    its base address, then try to grab any standard serial port
    address, and finally try to get any free port. */
    i = first_tuple(handle, &tuple, &parse);

    while (i != CS_NO_MORE_ITEMS) {
	if ((i == CS_SUCCESS) && (cf->io.nwin > 0) && ((cf->io.flags & CISTPL_IO_LINES_MASK) <= 3)) {
	    link->conf.ConfigIndex = cf->index;

	    for (j = 0; j < 5; j++) {
		link->io.BasePort1 = base[j];
		link->io.IOAddrLines = base[j] ? 16 : 3;

		i = CardServices(RequestIO, link->handle, &link->io);

		if (i == CS_SUCCESS)
		    goto found_port;
	    }
	}
    
	i = next_tuple(handle, &tuple, &parse);
    }

found_port:

    if (i != CS_SUCCESS) {
	printk(KERN_ERR "bt2000e_cs: no usable port range found\n");
	cs_error(link->handle, RequestIO, i);
	goto failed;
    }

    i = CardServices(RequestIRQ, link->handle, &link->irq);

    if (i != CS_SUCCESS) {
	cs_error(link->handle, RequestIRQ, i);
	link->irq.AssignedIRQ = 0;
    }

    i = CardServices(RequestConfiguration, link->handle, &link->conf);

    if (i != CS_SUCCESS) {
	cs_error(link->handle, RequestConfiguration, i);
	goto failed;
    }
  
    MOD_INC_USE_COUNT;

    if (bt2000e_open(info) != 0)
	goto failed;

    strcpy(info->node.dev_name, info->hdev.name);
    link->dev = &info->node;
    link->state &= ~DEV_CONFIG_PENDING;

    return;


cs_failed:

    cs_error(link->handle, last_fn, last_ret);

failed:

    bt2000e_release((u_long)link);
    link->state &= ~DEV_CONFIG_PENDING;

}


static void bt2000e_release(u_long arg) {

    dev_link_t *link = (dev_link_t *)arg;
    bt2000e_info_t *info = link->priv;

    if (link->state & DEV_PRESENT)
	bt2000e_close(info);

    MOD_DEC_USE_COUNT;

    link->dev = NULL;

    CardServices(ReleaseConfiguration, link->handle);
    CardServices(ReleaseIO, link->handle, &link->io);
    CardServices(ReleaseIRQ, link->handle, &link->irq);

    link->state &= ~DEV_CONFIG;

}


static int bt2000e_event(event_t event, int priority, event_callback_args_t *args) {

    dev_link_t *link = args->client_data;
    bt2000e_info_t *info = link->priv;
    unsigned long flags;
    
    switch (event) {
    
	case CS_EVENT_CARD_REMOVAL:
#if defined(OX950_DEBUG)
	    printk(KERN_DEBUG "bt2000e_cs: CS_EVENT_CARD_REMOVAL\n");
#endif
	    if (link->state & DEV_CONFIG) {
		bt2000e_release((u_long)link);
		mod_timer(&link->release, jiffies + HZ/20);
	    }
	    link->state &= ~DEV_PRESENT;
	    break;

	case CS_EVENT_CARD_INSERTION:
#if defined(OX950_DEBUG)
	    printk(KERN_DEBUG "bt2000e_cs: CS_EVENT_CARD_INSERTION\n");
#endif
	    link->state |= DEV_PRESENT | DEV_CONFIG_PENDING;
	    bt2000e_config(link);
	    break;

	case CS_EVENT_PM_SUSPEND:
#if defined(OX950_DEBUG)
	    printk(KERN_DEBUG "bt2000e_cs: CS_EVENT_PM_SUSPEND\n");
#endif
	    link->state |= DEV_SUSPEND;
	    /* Fall through... */

	case CS_EVENT_RESET_PHYSICAL:
#if defined(OX950_DEBUG)
	    printk(KERN_DEBUG "bt2000e_cs: CS_EVENT_RESET_PHYSICAL\n");
#endif
	    if (link->state & DEV_CONFIG) {
		tasklet_disable(&(info->tx_task));
		CardServices(ReleaseConfiguration, link->handle);
	    }
	    break;

	case CS_EVENT_PM_RESUME:
#if defined(OX950_DEBUG)
	    printk(KERN_DEBUG "bt2000e_cs: CS_EVENT_PM_RESUME\n");
#endif
	    link->state &= ~DEV_SUSPEND;
	    /* Fall through... */

	case CS_EVENT_CARD_RESET:
#if defined(OX950_DEBUG)
	    printk(KERN_DEBUG "bt2000e_cs: CS_EVENT_CARD_RESET\n");
#endif
	    if (link->state & DEV_CONFIG) {
		CardServices(RequestConfiguration, link->handle, &link->conf);
		write_lock_irqsave(&(info->lock), flags);
		bt2000e_stop_uart(info);
#if defined(OX950_CORRST)
		/* pcmcia COR reset */
		pcmcia_cor_reset(info);
#endif
		bt2000e_start_uart(info);
		write_unlock_irqrestore(&(info->lock), flags);
		/* reset Ericsson ROK module */
		bt2000e_hci_reset(info);
		/* FIXME: work-around against CTS problem, not really elegant, eh?
		   The problem is: we lose first TX because CTS is set low by ROK.		    
		*/
		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(HZ);
		/* ping ROK */
		bt2000e_tx_cts(info);
		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(HZ);
		/* we should be good to go now... */
		tasklet_enable(&(info->tx_task));
		tasklet_schedule(&(info->tx_task));
	    }
	    break;
    }

  return 0;

}



/* ======================== Module initialization ======================== */


static int __init init_bt2000e_cs(void) {

    servinfo_t serv;
    int err;

    printk(KERN_NOTICE "%s %s\n", DRIVER_NAME, DRIVER_VERSION);

    CardServices(GetCardServicesInfo, &serv);
  
    if (serv.Revision != CS_RELEASE_CODE) {
	printk(KERN_ERR "bt2000e_cs: Card Services release does not match!\n");
//	return -1;
    }

    err = register_pccard_driver(&dev_info, &bt2000e_attach, &bt2000e_detach);

    return err;

}


static void __exit exit_bt2000e_cs(void) {

    unregister_pccard_driver(&dev_info);

    while (dev_list != NULL)
	bt2000e_detach(dev_list);

}


module_init(init_bt2000e_cs);
module_exit(exit_bt2000e_cs);

EXPORT_NO_SYMBOLS;
