/*
 * tologo
 * Copyright (C) 2002 DindinX <David@dindinx.org>
 *
 * Most of this code is taken from the GIMP.
 * The GIMP -- an image manipulation program
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 * PNM reading and writing code Copyright (C) 1996 Erik Nygren
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
 * The pnm reading and writing code was written from scratch by Erik Nygren
 * (nygren@mit.edu) based on the specifications in the man pages and
 * does not contain any code from the netpbm or pbmplus distributions.
 */

#include <setjmp.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "tologo.h"

typedef struct _PNMScanner {
	int fd;			/* The file descriptor of the file being read */
	char cur;		/* The current character in the input stream */
	int eof;		/* Have we reached end of file? */
	char *inbuf;		/* Input buffer - initially 0 */
	int inbufsize;		/* Size of input buffer */
	int inbufvalidsize;	/* Size of input buffer with valid data */
	int inbufpos;		/* Position in input buffer */
} PNMScanner;

typedef struct _PNMInfo {
	int xres, yres;		/* The size of the image */
	int maxval;		/* For ascii image files, the max value
				 * which we need to normalize to */
	int np;			/* Number of image planes (0 for pbm) */
	int asciibody;		/* 1 if ascii body, 0 if raw body */
	jmp_buf jmpbuf;		/* Where to jump to on an error loading */
	/* Routine to use to load the pnm body */
	void (*loader) (PNMScanner *, struct _PNMInfo *, LogoStruct *);
} PNMInfo;

#define BUFLEN 512		/* The input buffer size for data returned
				 * from the scanner.  Note that lines
				 * aren't allowed to be over 256 characters
				 * by the spec anyways so this shouldn't
				 * be an issue. */

static void pnm_load_ascii(PNMScanner * scan,
			   PNMInfo * info, LogoStruct * pixel_rgn);
static void pnm_load_raw(PNMScanner * scan,
			 PNMInfo * info, LogoStruct * pixel_rgn);
static void pnm_load_rawpbm(PNMScanner * scan,
			    PNMInfo * info, LogoStruct * pixel_rgn);

static void pnmscanner_destroy(PNMScanner * s);
static void pnmscanner_createbuffer(PNMScanner * s, int bufsize);
static void pnmscanner_getchar(PNMScanner * s);
static void pnmscanner_eatwhitespace(PNMScanner * s);
static void pnmscanner_gettoken(PNMScanner * s, char *buf, int bufsize);
static void pnmscanner_getsmalltoken(PNMScanner * s, char *buf);

static PNMScanner *pnmscanner_create(int fd);

#define pnmscanner_eof(s) ((s)->eof)
#define pnmscanner_fd(s) ((s)->fd)

/* Checks for a fatal error */
#define CHECK_FOR_ERROR(predicate, jmpbuf, errmsg) \
        if ((predicate)) \
        { fprintf(stderr, (errmsg)); longjmp((jmpbuf),1); }

static struct struct_pnm_types {
	char name;
	int np;
	int asciibody;
	int maxval;
	void (*loader) (PNMScanner *, struct _PNMInfo *,
			LogoStruct * pixel_rgn);
} pnm_types[] = {
	{
	'1', 0, 1, 1, pnm_load_ascii},	/* ASCII PBM */
	{
	'2', 1, 1, 255, pnm_load_ascii},	/* ASCII PGM */
	{
	'3', 3, 1, 255, pnm_load_ascii},	/* ASCII PPM */
	{
	'4', 0, 0, 1, pnm_load_rawpbm},	/* RAW   PBM */
	{
	'5', 1, 0, 255, pnm_load_raw},	/* RAW   PGM */
	{
	'6', 3, 0, 255, pnm_load_raw},	/* RAW   PPM */
	{
	0, 0, 0, 0, NULL}
};

LogoStruct *
load_image(char *filename)
{
	LogoStruct *pixel_rgn = NULL;
	int fd;			/* File descriptor */
	char buf[BUFLEN];	/* buffer for random things like scanning */
	PNMInfo *pnminfo;
	PNMScanner *volatile scan;
	int ctr;

	printf("Loading %s:\n", filename);

	/* open the file */
	fd = open(filename, O_RDONLY);

	if (fd == -1) {
		fprintf(stderr, "PNM: Can't open file %s.", filename);
		return NULL;
	}

	/* allocate the necessary structures */
	pnminfo = malloc(sizeof (PNMInfo));

	pixel_rgn = malloc(sizeof (LogoStruct));
	scan = NULL;
	/* set error handling */
	if (setjmp(pnminfo->jmpbuf)) {
		/* If we get here, we had a problem reading the file */
		if (scan)
			pnmscanner_destroy(scan);
		close(fd);
		free(pnminfo);
		if (pixel_rgn)
			free(pixel_rgn);
		return NULL;
	}

	if (!(scan = pnmscanner_create(fd)))
		longjmp(pnminfo->jmpbuf, 1);

	/* Get magic number */
	pnmscanner_gettoken(scan, buf, BUFLEN);
	CHECK_FOR_ERROR(pnmscanner_eof(scan), pnminfo->jmpbuf,
			"PNM: Premature end of file.");
	CHECK_FOR_ERROR((buf[0] != 'P' || buf[2]), pnminfo->jmpbuf,
			"PNM: Invalid file.");

	printf("magic number gotten: P%c\n", buf[1]);
	/* Look up magic number to see what type of PNM this is */
	for (ctr = 0; pnm_types[ctr].name; ctr++)
		if (buf[1] == pnm_types[ctr].name) {
			pnminfo->np = pnm_types[ctr].np;
			pnminfo->asciibody = pnm_types[ctr].asciibody;
			pnminfo->maxval = pnm_types[ctr].maxval;
			pnminfo->loader = pnm_types[ctr].loader;
		}
	if (!pnminfo->loader) {
		fprintf(stderr, "PNM: File not in a supported format.");
		longjmp(pnminfo->jmpbuf, 1);
	}

	pnmscanner_gettoken(scan, buf, BUFLEN);
	CHECK_FOR_ERROR(pnmscanner_eof(scan), pnminfo->jmpbuf,
			"PNM: Premature end of file.");
	pnminfo->xres = isdigit(*buf) ? atoi(buf) : 0;
	CHECK_FOR_ERROR(pnminfo->xres <= 0, pnminfo->jmpbuf,
			"PNM: Invalid X resolution.");

	pnmscanner_gettoken(scan, buf, BUFLEN);
	CHECK_FOR_ERROR(pnmscanner_eof(scan), pnminfo->jmpbuf,
			"PNM: Premature end of file.");
	pnminfo->yres = isdigit(*buf) ? atoi(buf) : 0;
	CHECK_FOR_ERROR(pnminfo->yres <= 0, pnminfo->jmpbuf,
			"PNM: Invalid Y resolution.");

	printf("res: (%d, %d)\n", pnminfo->xres, pnminfo->yres);
	pixel_rgn->width = pnminfo->xres;
	pixel_rgn->height = pnminfo->yres;
	pixel_rgn->rgb_data = malloc(3 * pixel_rgn->width * pixel_rgn->height);
	pixel_rgn->image = malloc(pixel_rgn->width * pixel_rgn->height);
	memset(pixel_rgn->rgb_data, 0,
	       3 * pixel_rgn->width * pixel_rgn->height);
	if (pnminfo->np != 0) {	/* pbm's don't have a maxval field */
		pnmscanner_gettoken(scan, buf, BUFLEN);
		CHECK_FOR_ERROR(pnmscanner_eof(scan), pnminfo->jmpbuf,
				"PNM: Premature end of file.");

		pnminfo->maxval = isdigit(*buf) ? atoi(buf) : 0;
		CHECK_FOR_ERROR(((pnminfo->maxval <= 0)
				 || (pnminfo->maxval > 255
				     && !pnminfo->asciibody)), pnminfo->jmpbuf,
				"PNM: Invalid maximum value.");
	}

	pnminfo->loader(scan, pnminfo, pixel_rgn);

	/* Destroy the scanner */
	pnmscanner_destroy(scan);

	/* free the structures */
	free(pnminfo);

	/* close the file */
	close(fd);

	return pixel_rgn;
}

static void
pnm_load_ascii(PNMScanner * scan, PNMInfo * info, LogoStruct * pixel_rgn)
{
	unsigned char *d;
	int x, y, b;
	int np;
	char buf[BUFLEN];

	np = (info->np) ? (info->np) : 1;

	/* Buffer reads to increase performance */
	pnmscanner_createbuffer(scan, 4096);

	for (y = 0; y < info->yres; y++) {
		d = pixel_rgn->rgb_data + 3 * y * pixel_rgn->width;

		for (x = 0; x < info->xres; x++) {
			for (b = 0; b < np; b++) {
				/* Truncated files will just have all 0's at the end of the images */
				CHECK_FOR_ERROR(pnmscanner_eof(scan),
						info->jmpbuf,
						"PNM: Premature end of file.");
				if (info->np)
					pnmscanner_gettoken(scan, buf, BUFLEN);
				else
					pnmscanner_getsmalltoken(scan, buf);
				switch (info->maxval) {
				case 255:
					d[b] = isdigit(*buf) ? atoi(buf) : 0;
					break;
				case 1:
					d[b] = (*buf == '0') ? 0xff : 0x00;
					break;
				default:
					d[b] = (unsigned char)
						(255.0 * (((double) (isdigit(*buf) ? atoi(buf) : 0))
							      / (double) (info-> maxval)));
				}
			}
			d += np;
		}
	}
}

static void
pnm_load_raw(PNMScanner * scan, PNMInfo * info, LogoStruct * pixel_rgn)
{
	unsigned char *d;
	int x, y;
	int fd;

	printf("loading raw pnm (maxval=%d)\n", info->maxval);
	fd = pnmscanner_fd(scan);

	for (y = 0; y < info->yres; y++) {
		d = pixel_rgn->rgb_data + 3 * y * pixel_rgn->width;

		CHECK_FOR_ERROR((info->xres * info->np
				 != read(fd, d, info->xres * info->np)),
				info->jmpbuf, "PNM: Premature end of file.");

		if (info->maxval != 255) {	/* Normalize if needed */
			for (x = 0; x < info->xres * info->np; x++)
				d[x] =
				    (unsigned char) (255.0 * (double) (d[x]) /
						     (double) (info->maxval));
		}
	}
}

static void
pnm_load_rawpbm(PNMScanner * scan, PNMInfo * info, LogoStruct * pixel_rgn)
{
	unsigned char *buf;
	unsigned char curbyte;
	unsigned char *d;
	int x, y;
	int fd;
	int rowlen, bufpos;

	fd = pnmscanner_fd(scan);
	rowlen = info->xres / 8;
	if (info->xres % 8)
		rowlen++;
	buf = malloc(sizeof (unsigned char) * rowlen);

	for (y = 0; y < info->yres; y++) {
		d = pixel_rgn->rgb_data + 3 * y * pixel_rgn->width;

		CHECK_FOR_ERROR((rowlen != read(fd, buf, rowlen)),
				info->jmpbuf, "PNM: Error reading file.");
		bufpos = 0;
		curbyte = buf[0];

		for (x = 0; x < info->xres; x++) {
			if ((x % 8) == 0)
				curbyte = buf[bufpos++];
			d[3 * x] = (curbyte & 0x80) ? 0x00 : 0xff;
			d[3 * x + 1] = d[3 * x];
			d[3 * x + 2] = d[3 * x];
			curbyte <<= 1;
		}
	}
	free(buf);
}

/**************** FILE SCANNER UTILITIES **************/

/* pnmscanner_create ---
 *    Creates a new scanner based on a file descriptor.  The
 *    look ahead buffer is one character initially.
 */
static PNMScanner *
pnmscanner_create(int fd)
{
	PNMScanner *s;

	s = malloc(sizeof (PNMScanner));
	s->fd = fd;
	s->inbuf = 0;
	s->eof = !read(s->fd, &(s->cur), 1);
	return (s);
}

/* pnmscanner_destroy ---
 *    Destroys a scanner and its resources.  Doesn't close the fd.
 */
static void
pnmscanner_destroy(PNMScanner * s)
{
	if (s->inbuf)
		free(s->inbuf);
	free(s);
}

/* pnmscanner_createbuffer ---
 *    Creates a buffer so we can do buffered reads.
 */
static void
pnmscanner_createbuffer(PNMScanner * s, int bufsize)
{
	s->inbuf = malloc(sizeof (char) * bufsize);
	s->inbufsize = bufsize;
	s->inbufpos = 0;
	s->inbufvalidsize = read(s->fd, s->inbuf, bufsize);
}

/* pnmscanner_gettoken ---
 *    Gets the next token, eating any leading whitespace.
 */
static void
pnmscanner_gettoken(PNMScanner * s, char *buf, int bufsize)
{
	int ctr = 0;

	pnmscanner_eatwhitespace(s);
	while (!(s->eof) && !isspace(s->cur) && (s->cur != '#')
	       && (ctr < bufsize)) {
		buf[ctr++] = s->cur;
		pnmscanner_getchar(s);
	}
	buf[ctr] = '\0';
}

/* pnmscanner_getsmalltoken ---
 *    Gets the next char, eating any leading whitespace.
 */
static void
pnmscanner_getsmalltoken(PNMScanner * s, char *buf)
{
	pnmscanner_eatwhitespace(s);
	if (!(s->eof) && !isspace(s->cur) && (s->cur != '#')) {
		*buf = s->cur;
		pnmscanner_getchar(s);
	}
}

/* pnmscanner_getchar ---
 *    Reads a character from the input stream
 */
static void
pnmscanner_getchar(PNMScanner * s)
{
	if (s->inbuf) {
		s->cur = s->inbuf[s->inbufpos++];
		if (s->inbufpos >= s->inbufvalidsize) {
			if (s->inbufsize > s->inbufvalidsize)
				s->eof = 1;
			else
				s->inbufvalidsize =
				    read(s->fd, s->inbuf, s->inbufsize);
			s->inbufpos = 0;
		}
	} else
		s->eof = !read(s->fd, &(s->cur), 1);
}

/* pnmscanner_eatwhitespace ---
 *    Eats up whitespace from the input and returns when done or eof.
 *    Also deals with comments.
 */
static void
pnmscanner_eatwhitespace(PNMScanner * s)
{
	int state = 0;

	while (!(s->eof) && (state != -1)) {
		switch (state) {
		case 0:	/* in whitespace */
			if (s->cur == '#') {
				state = 1;	/* goto comment */
				pnmscanner_getchar(s);
			} else if (!isspace(s->cur))
				state = -1;
			else
				pnmscanner_getchar(s);
			break;

		case 1:	/* in comment */
			if (s->cur == '\n')
				state = 0;	/* goto whitespace */
			pnmscanner_getchar(s);
			break;
		}
	}
}
